import boto3
import click
import os
import re

from functools import reduce
from rireki.core.store import Store


class AmazonWebServices(Store):
    NAME = 'aws'

    def __init__(self):
        Store.__init__(self)

        self.region = None
        self.access_key = None
        self.access_secret = None
        self.bucket = None
        self.path = None

    def ask_config(self):
        Store.ask_config(self)

        click.echo('Please, introduce credentials for %s.' % self._get_service_name())

        self.region = self.__ask_region()
        self.access_key = self.__ask_access_key()
        self.access_secret = self.__ask_access_secret()
        self.bucket = self.__ask_bucket()
        self.path = self.__ask_path()

    def load_config(self, config):
        Store.load_config(self, config)

        self.region = config['region']
        self.access_key = config['access_key']
        self.access_secret = config['access_secret']
        self.bucket = config['bucket']
        self.path = config['path']

    def get_config(self):
        config = Store.get_config(self)

        config['region'] = self.region
        config['access_key'] = self.access_key
        config['access_secret'] = self.access_secret
        config['bucket'] = self.bucket
        config['path'] = self.path

        return config

    def _get_backup_names(self):
        client = self.__create_s3_client()
        paginator = client.get_paginator('list_objects')
        page_iterator = paginator.paginate(Bucket=self.bucket, Prefix=self.path)

        names = reduce(
            sum,
            [self.__parse_page_backups(self.path, page) for page in page_iterator],
        )

        return sorted(names, reverse=True)

    def _upload_file(self, source, destination):
        client = self.__create_s3_client()

        client.upload_file(
            source,
            self.bucket,
            os.path.join(self.path, destination),
        )

    def _get_service_name(self):
        return 'Amazon Web Services'

    def _get_endpoint_url(self):
        return 'https://s3.%s.amazonaws.com' % self.region

    def __ask_region(self):
        return click.prompt('Region')

    def __ask_access_key(self):
        return click.prompt('Access Key')

    def __ask_access_secret(self):
        return click.prompt('Access Secret (hidden)', hide_input=True)

    def __ask_bucket(self):
        return click.prompt('Bucket')

    def __ask_path(self):
        return click.prompt('Path')

    def __create_s3_client(self):
        return boto3.client(
            's3',
            region_name=self.region,
            endpoint_url=self._get_endpoint_url(),
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.access_secret,
        )

    def __parse_page_backups(self, path, page):
        backup_names = set()

        if 'Contents' not in page:
            return backup_names

        if not path.endswith('/'):
            path += '/'

        regex = '%s([^\\/.]+).*$' % path.replace('/', '\\/')

        for item in page['Contents']:
            match = re.match(regex, item['Key'])

            if match:
                backup_names.add(match.group(1))

        return list(backup_names)



import os

from rireki.testing.cli import Cli
from rireki.testing.test_case import TestCase
from rireki.utils.file_helpers import touch
from rireki.utils.time_helpers import now, DAY_SECONDS, YEAR_SECONDS


class TestBackup(TestCase):

    def test_without_installed_projects(self):
        # Execute
        result = Cli.run('clean')

        # Assert
        assert result.exit_code == 0
        assert 'No projects installed!' in result.output

    def test_without_stale_backups(self):
        # Prepare
        project = self._create_project(
            store='local',
            store_config={'path': '/tmp/rireki_testing/store'},
        )

        # Execute
        result = Cli.run('clean')

        # Assert
        assert result.exit_code == 0
        assert ('Project "%s" does not have any stale backups' % project.name) in result.output
        assert 'Done' in result.output
        assert 'Error' not in result.output

    def test_with_stale_backups(self):
        # Prepare
        project = self._create_project(
            store='local',
            store_config={'path': '/tmp/rireki_testing/store'},
            retention={'last_backups_retention': 1, 'year_backups_retention': 'monthly'}
        )

        today = now()
        yesterday = today - DAY_SECONDS
        last_month = today - YEAR_SECONDS

        touch('/tmp/rireki_testing/store/%s/backup' % today)
        touch('/tmp/rireki_testing/store/%s/backup' % yesterday)
        touch('/tmp/rireki_testing/store/%s/backup' % last_month)

        # Execute
        result = Cli.run('clean')

        # Assert
        assert result.exit_code == 0
        assert ('Cleaning up %s...' % project.name) in result.output
        assert str(yesterday) in result.output
        assert 'Done' in result.output
        assert 'Error' not in result.output

        assert os.path.exists('/tmp/rireki_testing/store/%s/backup' % today)
        assert not os.path.exists('/tmp/rireki_testing/store/%s/backup' % yesterday)
        assert os.path.exists('/tmp/rireki_testing/store/%s/backup' % last_month)



import click

from rireki.commands.add import add
from rireki.commands.backup import backup
from rireki.commands.clean import clean
from rireki.commands.status import status
from rireki.core.config import Config


@click.group()
def cli():
    Config.load()


cli.add_command(add)
cli.add_command(backup)
cli.add_command(clean)
cli.add_command(status)

if __name__ == '__main__':
    cli()



import os

from rireki.core.backup import Backup
from rireki.core.configurable import Configurable
from rireki.utils.file_helpers import file_get_extension
from rireki.utils.output import format_time
from rireki.utils.time_helpers import now


class Store(Configurable):

    def __init__(self):
        Configurable.__init__(self)

        self.project = None

    def create_backup(self, files_path):
        backup_name = '{slug}-backup-{date}-{timestamp}'.format(
            slug=self.project.slug,
            date=format_time(now(), 'date'),
            timestamp=now(),
        )

        if os.path.isfile(files_path):
            self._upload_file(
                files_path,
                '{}.{}'.format(backup_name, file_get_extension(files_path)),
            )
        elif os.path.isdir(files_path):
            for file in os.listdir(files_path):
                self._upload_file(
                    os.path.join(files_path, file),
                    os.path.join(backup_name, file),
                )

    def remove_backup(self, backup):
        raise Exception('%s store must implement remove_backup method' % self.name)

    def get_last_backup(self):
        backups = self.get_backups()

        if not backups:
            return None

        return backups[0]

    def get_backups(self):
        names = self._get_backup_names()

        return [Backup(name) for name in names if Backup.is_backup_name(name)]

    def _get_backup_names(self):
        raise Exception('%s store must implement _get_backup_names method' % self.name)

    def _upload_file(self, source, destination):
        raise Exception('%s store must implement _upload_file method' % self.name)



import os


class Config():
    home_path = None
    projects_path = None

    @classmethod
    def load(cls):
        cls.home_path = os.environ.get('RIREKI_HOME') or cls.__default_home()
        cls.projects_path = cls.home_path + '/projects'

    @classmethod
    def __default_home(cls):
        return '%s/.rireki' % os.path.expanduser('~')



import re


def str_studly(text):
    words = text.split(' ')

    return ''.join([w[0].upper() + w[1:].lower() for w in words])


def str_pad(text, length, padding=' '):
    return (text + (padding * length))[:length]


def str_slug(text):
    text = text.lower()

    text = re.sub('\\s+', '-', text)
    text = re.sub('[^\\w-]+', '', text)

    return text



from rireki.core.configurable import Configurable
from rireki.utils.string_helpers import str_slug
from rireki.utils.time_helpers import YEAR_SECONDS
from rireki.core.retention_policy import RetentionPolicy


class Project(Configurable):

    def __init__(self, name, driver, store):
        Configurable.__init__(self, name)

        self.last_backups_retention = 7
        self.year_backups_retention = RetentionPolicy.MONTHLY
        self.ancient_backups_retention = RetentionPolicy.YEARLY
        self.driver = driver
        self.store = store

    @property
    def slug(self):
        return str_slug(self.name)

    def load_config(self, config):
        Configurable.load_config(self, config)

        year_retention = RetentionPolicy.from_string(config.get('year_backups_retention'))
        ancient_retention = RetentionPolicy.from_string(config.get('ancient_backups_retention'))

        self.last_backups_retention = config.get('last_backups_retention') or self.last_backups_retention
        self.year_backups_retention = year_retention or self.year_backups_retention
        self.ancient_backups_retention = ancient_retention or self.ancient_backups_retention

        self.driver.project = self
        self.store.project = self

        self.driver.load_config(config['driver'])
        self.store.load_config(config['store'])

    def get_config(self):
        config = Configurable.get_config(self)

        config['last_backups_retention'] = self.last_backups_retention
        config['year_backups_retention'] = self.year_backups_retention.value
        config['ancient_backups_retention'] = self.ancient_backups_retention.value
        config['driver'] = self.driver.get_config()
        config['store'] = self.store.get_config()

        return config

    def has_pending_backups(self):
        last_backup = self.get_last_backup()

        if not last_backup:
            return True

        return self.driver.has_pending_backups(last_backup.time)

    def get_backups(self):
        return self.store.get_backups()

    def get_last_backup(self):
        return self.store.get_last_backup()

    def get_stale_backups(self):
        backups = sorted(self.get_backups(), key=lambda x: x.time, reverse=True)
        year_retention = self.year_backups_retention
        ancient_retention = self.ancient_backups_retention
        retained_backups = []
        stale_backups = []

        for backup in backups:
            previous_backup = retained_backups[-1] if retained_backups else None
            retention_policy = year_retention if backup.get_age() < YEAR_SECONDS else ancient_retention

            if len(retained_backups) < self.last_backups_retention or backup.retain(previous_backup, retention_policy):
                retained_backups.append(backup)
                continue

            stale_backups.append(backup)

        return stale_backups

    def perform_backup(self):
        self.driver.perform_backup()

    def remove_backup(self, backup):
        self.store.remove_backup(backup)



class Configurable():

    def __init__(self, name=None):
        self.name = name or self.NAME

    def ask_config(self):
        pass

    def load_config(self, config):
        pass

    def get_config(self):
        return {'name': self.name}



import re
from datetime import datetime

from rireki.utils.time_helpers import now
from rireki.core.retention_policy import RetentionPolicy


class Backup(object):

    @classmethod
    def is_backup_name(cls, name):
        return re.search('(\\d+)$', name)

    def __init__(self, name):
        self.name = name

    @property
    def time(self):
        matches = re.findall('(\\d+)$', self.name)

        return int(matches[-1]) if matches else 0

    def get_age(self):
        return now() - self.time

    def retain(self, previous_backup, retention_policy):
        if retention_policy == RetentionPolicy.NONE:
            return False

        if not previous_backup:
            return True

        backup_date = datetime.fromtimestamp(self.time)
        previous_date = datetime.fromtimestamp(previous_backup.time)

        if retention_policy == RetentionPolicy.YEARLY:
            return backup_date.year != previous_date.year

        if retention_policy == RetentionPolicy.MONTHLY:
            return (backup_date.year, backup_date.month) != (previous_date.year, previous_date.month)

        if retention_policy == RetentionPolicy.WEEKLY:
            backup_year, backup_week = backup_date.isocalendar()[:2]
            previous_year, previous_week = previous_date.isocalendar()[:2]
            return (backup_year, backup_week) != (previous_year, previous_week)

        return True



import click
import os

from rireki.core.store import Store
from rireki.utils.file_helpers import file_get_name
from shutil import copyfile, rmtree


class Local(Store):
    NAME = 'local'

    def __init__(self):
        Store.__init__(self)

        self.path = None

    def ask_config(self):
        Store.ask_config(self)

        self.path = self.__ask_path()

    def load_config(self, config):
        Store.load_config(self, config)

        self.path = config['path']

    def get_config(self):
        config = Store.get_config(self)

        config['path'] = self.path

        return config

    def remove_backup(self, backup):
        rmtree(os.path.join(self.path, backup.name))

    def _get_backup_names(self):
        if not os.path.exists(self.path):
            return []

        names = []

        for file in os.listdir(self.path):
            names.append(file_get_name(file))

        return sorted(names, reverse=True)

    def _upload_file(self, source, destination):
        destination = os.path.join(self.path, destination)
        destination_parent = os.path.dirname(destination)

        if not os.path.exists(destination_parent):
            os.makedirs(destination_parent)

        copyfile(source, destination)

    def __ask_path(self):
        return click.prompt('Where do you want to store the backup files?')



import os
import toml

from rireki.core.config import Config
from rireki.core.project import Project
from rireki.drivers.index import drivers
from rireki.stores.index import stores


class ProjectsManager():

    @classmethod
    def get_projects(cls):
        if not os.path.exists(Config.projects_path):
            return []

        return [cls.__parse_project_config(file_name[:-5]) for file_name in os.listdir(Config.projects_path)]

    @classmethod
    def get_project_by_name(cls, name):
        if not cls.project_exists(name):
            return None

        return cls.__parse_project_config(name)

    @classmethod
    def project_exists(cls, name):
        return os.path.exists('%s/%s.conf' % (Config.projects_path, name))

    @classmethod
    def install_project(cls, project):
        if not os.path.exists(Config.projects_path):
            os.makedirs(Config.projects_path)

        with open('%s/%s.conf' % (Config.projects_path, project.name), 'w') as config_file:
            config_file.write(toml.dumps(project.get_config()))

    @classmethod
    def __parse_project_config(cls, project_name):
        config = toml.load('%s/%s.conf' % (Config.projects_path, project_name))

        driver = drivers[config['driver']['name']]()
        store = stores[config['store']['name']]()
        project = Project(config['name'], driver, store)

        project.load_config(config)

        return project



from rireki.testing.test_case import TestCase
from rireki.utils.string_helpers import str_slug


class TestStringHelpers(TestCase):

    def test_slug(self):
        assert str_slug('Miyamoto Musashi') == 'miyamoto-musashi'
        assert str_slug('Let\'s Go') == 'lets-go'



import os
import re

from rireki.testing.cli import Cli
from rireki.testing.test_case import TestCase
from rireki.utils.file_helpers import touch
from rireki.utils.time_helpers import now


class TestStatus(TestCase):

    def test_without_installed_projects(self):
        # Execute
        result = Cli.run('status')

        # Assert
        assert result.exit_code == 0
        assert 'No projects installed!' in result.output

        assert not os.path.exists(self.home_path)

    def test_with_one_project_with_backups_pending(self):
        # Prepare
        project = self._create_project(driver='files', store='local')

        # Execute
        result = Cli.run('status')

        # Assert
        assert result.exit_code == 0

        output_lines = result.output.splitlines()
        assert len(output_lines) == 2
        assert project.name in output_lines[1]
        assert 'files' in output_lines[1]
        assert 'local' in output_lines[1]
        assert 'backup-pending' in output_lines[1]

    def test_with_one_project_backed_up(self):
        # Prepare
        project = self._create_project(
            store='local',
            store_config={'path': '/tmp/rireki_testing/store'},
        )

        touch('/tmp/rireki_testing/store/%s/backup' % now())

        # Execute
        result = Cli.run('status')

        # Assert
        assert result.exit_code == 0

        output_lines = result.output.splitlines()
        assert len(output_lines) == 2
        assert project.name in output_lines[1]
        assert re.search('Backed up \\d seconds ago', output_lines[1])



import click

from rireki.core.projects_manager import ProjectsManager


@click.command()
@click.argument('project', required=False)
def clean(project=None):
    """Clean stale backups"""

    if project:
        name = project
        project = ProjectsManager.get_project_by_name(name)

        if not project:
            click.echo('Project with name "%s" is not installed!' % name)
            return

        projects = [project]
    else:
        projects = ProjectsManager.get_projects()

    if not projects:
        click.echo('No projects installed!')
        return

    for project in projects:
        __process_cleanup(project)

    click.echo('Done!')


def __process_cleanup(project):
    stale_backups = project.get_stale_backups()

    if not stale_backups:
        click.echo('Project "%s" does not have any stale backups' % project.name)
        return

    click.echo('Cleaning up %s...' % project.name)

    try:
        for stale_backup in stale_backups:
            click.echo('Removing %s...' % stale_backup.name)

            project.remove_backup(stale_backup)
    except Exception as e:
        error_message = click.style('Error: %s' % e, fg='red')

        click.echo(error_message, err=True)



import os

from rireki.core.project import Project
from rireki.drivers.files import Files
from rireki.stores.local import Local
from rireki.testing.test_case import TestCase
from rireki.utils.file_helpers import touch
from rireki.utils.string_helpers import str_slug
from rireki.utils.time_helpers import now


class TestFiles(TestCase):

    def setUp(self):
        TestCase.setUp(self)

        self.driver = Files()
        self.store = Local()
        self.project = Project(self.faker.name(), self.driver, self.store)

        self.driver.project = self.project
        self.store.project = self.project

    def test_creates_backups_with_one_path(self):
        # Prepare
        tmp_path = os.path.join(self.home_path, '{}-{}'.format(self.project.slug, now()))
        store_path = os.path.join(tmp_path, 'backups')
        driver_paths = [os.path.join(tmp_path, 'files')]

        self.store.load_config({'path': store_path})
        self.driver.load_config({
            'frequency': 42,
            'paths': driver_paths,
        })

        touch(os.path.join(driver_paths[0], str_slug(self.faker.word())))

        # Execute
        self.driver.perform_backup()

        # Assert
        assert os.path.exists(store_path)

        backup = self.store.get_last_backup()
        assert backup is not None

        assert os.path.exists(os.path.join(store_path, backup.name + '.zip'))

    def test_creates_backups_with_multiple_paths(self):
        # Prepare
        tmp_path = os.path.join(self.home_path, '{}-{}'.format(self.project.slug, now()))
        store_path = os.path.join(tmp_path, 'backups')
        driver_paths = [
            os.path.join(tmp_path, 'files-1'),
            os.path.join(tmp_path, 'files-2'),
        ]

        self.store.load_config({'path': store_path})
        self.driver.load_config({
            'frequency': 42,
            'paths': driver_paths,
        })

        touch(os.path.join(driver_paths[0], str_slug(self.faker.word())))
        touch(os.path.join(driver_paths[1], str_slug(self.faker.word())))

        # Execute
        self.driver.perform_backup()

        # Assert
        assert os.path.exists(store_path)

        backup = self.store.get_last_backup()
        assert backup is not None

        assert os.path.exists(os.path.join(store_path, backup.name + '.zip'))

    def test_creates_backups_with_files(self):
        # Prepare
        tmp_path = os.path.join(self.home_path, '{}-{}'.format(self.project.slug, now()))
        file_name = str_slug(self.faker.word())
        store_path = os.path.join(tmp_path, 'backups')
        driver_paths = [os.path.join(tmp_path, 'files', file_name)]

        self.store.load_config({'path': store_path})
        self.driver.load_config({
            'frequency': 42,
            'paths': driver_paths,
        })

        touch(driver_paths[0])

        # Execute
        self.driver.perform_backup()

        # Assert
        assert os.path.exists(store_path)

        backup = self.store.get_last_backup()
        assert backup is not None

        assert os.path.exists(os.path.join(store_path, backup.name + '.zip'))



from rireki.testing.test_case import TestCase
from rireki.utils.output import format_time


class TestOutput(TestCase):

    def test_format_time_interval(self):
        assert format_time(1, 'interval') == '1 second'
        assert format_time(42, 'interval') == '42 seconds'
        assert format_time(61, 'interval') == '1 minute'
        assert format_time(150, 'interval') == '2 minutes'
        assert format_time(3601, 'interval') == '1 hour'
        assert format_time(14423, 'interval') == '4 hours'
        assert format_time(86412, 'interval') == '1 day'
        assert format_time(259211, 'interval') == '3 days'
        assert format_time(3628811, 'interval') == '42 days'



from enum import Enum


class RetentionPolicy(Enum):
    NONE = 'none'
    WEEKLY = 'weekly'
    MONTHLY = 'monthly'
    YEARLY = 'yearly'

    @classmethod
    def from_string(cls, value):
        try:
            return cls(value)
        except ValueError:
            return None



import click
import json
import os
import subprocess

from rireki.core.driver import Driver
from rireki.utils.file_helpers import file_put_contents
from threading import Thread


class Custom(Driver):
    NAME = 'custom'

    def __init__(self):
        Driver.__init__(self)

        self.command = None
        self.timeout = 60

    def ask_config(self):
        Driver.ask_config(self)

        self.command = self.__ask_command()

    def load_config(self, config):
        Driver.load_config(self, config)

        self.command = config['command']
        self.timeout = config['timeout']

    def get_config(self):
        config = Driver.get_config(self)

        config['timeout'] = self.timeout
        config['command'] = self.command

        return config

    def _prepare_backup_files(self, path):
        logs = self.__run_command(path)

        file_put_contents(os.path.join(path, 'logs.json'), json.dumps(logs))

        return path

    def __ask_command(self):
        return click.prompt('Enter the command you want to execute to perform backups')

    def __run_command(self, path):
        self.process = None
        self.logs = {}

        def target():
            self.process = subprocess.Popen(
                self.command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env={'RIREKI_BACKUP_PATH': path}
            )

            self.logs['stdout'] = str(self.process.stdout.read())
            self.logs['stderr'] = str(self.process.stderr.read())

        thread = Thread(target=target)

        thread.start()
        thread.join(self.timeout)
        self.process.poll()

        if thread.is_alive() or self.process.returncode is None:
            self.process.terminate()
            thread.join()
            self.process.poll()

        if self.process.returncode != 0:
            raise Exception(
                'Command failed with return code %s\n\nstdout:\n%s\nstderr:\n%s' %
                (self.process.returncode, self.logs['stdout'], self.logs['stderr'])
            )

        return self.logs



import click

from datetime import datetime
from rireki.utils.array_helpers import array_map
from rireki.utils.string_helpers import str_pad


def display_table(headers, rows, min_column_width=14):
    headers = array_map(lambda h: h.upper(), headers)

    column_widths = __calculate_column_widths(headers, rows, min_column_width)

    click.echo(__format_table_row(headers, column_widths))

    for row in rows:
        click.echo(__format_table_row(row, column_widths))


def format_time(time, format):
    if format == 'date':
        return __format_time_date(time)
    elif format == 'interval':
        return __format_time_interval(time)


def __calculate_column_widths(headers, rows, min_column_width):
    column_widths = [min_column_width] * len(headers)

    column_widths = __fit_row_column_widths(headers, column_widths)

    for row in rows:
        columns_content = array_map(lambda cell: __render_cell_content(cell), row)
        column_widths = __fit_row_column_widths(columns_content, column_widths)

    return column_widths


def __fit_row_column_widths(columns_content, column_widths):
    for index, content in enumerate(columns_content):
        column_widths[index] = max(column_widths[index], len(content))

    return column_widths


def __format_time_interval(time):
    seconds = float(time)
    if seconds <= 60:
        return __format_time_unit(seconds, 'second')

    minutes = seconds / 60
    if minutes <= 60:
        return __format_time_unit(minutes, 'minute')

    hours = minutes / 60
    if hours <= 24:
        return __format_time_unit(hours, 'hour')

    days = hours / 24
    return __format_time_unit(days, 'day')


def __format_time_date(time):
    return datetime.fromtimestamp(time).strftime('%Y-%m-%d')


def __format_time_unit(magnitude, name):
    magnitude = int(magnitude)

    if not magnitude == 1:
        name = name + 's'

    return '%s %s' % (str(magnitude), name)


def __format_table_row(row, column_widths):
    text = ''

    for index, cell_content in enumerate(row):
        text = text + __format_table_cell(cell_content, column_widths[index])

    return text


def __format_table_cell(content, column_width):
    if type(content) is not dict:
        return str_pad(content, column_width)

    text = str_pad(content['text'], column_width)

    if 'color' in content:
        text = click.style(text, fg=content['color'])

    return text


def __render_cell_content(content):
    if type(content) is not dict:
        return content

    return content['text']



from time import time

testing_now = None
DAY_SECONDS = 60 * 60
YEAR_SECONDS = 365 * 24 * DAY_SECONDS


def now():
    global testing_now

    return testing_now or int(time())


def set_testing_now(time):
    global testing_now

    testing_now = time



from mock import Mock
from rireki.core.project import Project
from rireki.drivers.custom import Custom
from rireki.testing.test_case import TestCase


class TestCustom(TestCase):

    def setUp(self):
        TestCase.setUp(self)

        self.driver = Custom()
        self.driver.project = Project(self.faker.name(), Mock(), Mock())

    # Golden path tested at rireki/commands/backup_test.py

    def test_timeout(self):
        # Prepare
        self.driver.load_config({
            'frequency': 42,
            'command': 'while true; do echo; done',
            'timeout': 1,
        })

        # Execute
        with self.assertRaises(Exception):
            self.driver.perform_backup()

        # Assert
        self.driver.project.store.create_backup.assert_not_called()

    def test_error(self):
        # Prepare
        self.driver.load_config({
            'frequency': 42,
            'command': 'exit 1',
            'timeout': 1,
        })

        # Execute
        with self.assertRaises(Exception):
            self.driver.perform_backup()

        # Assert
        self.driver.project.store.create_backup.assert_not_called()
